from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class StaticModelsConfig(AppConfig):
    name = 'static_models'
    verbose_name = _("Generate static pages from models")



from django.conf import settings as dsettings
from django.core.exceptions import ImproperlyConfigured
from django.utils import module_loading


#! Generally deprecated, in favour of a few checks
class Settings():
    '''
    Gather settings from the settings file.
    For easy access and maybe caching. Django enables attribute access 
    on the Settings.py file, but it is awkward poking in for app-related
    settings.
    
    app_dirs 
        if true, seek for image filters in registtered apps
    modules 
        module paths to seek image_filters modules 
    '''
    def __init__(self):
        # defaults
        self.STATICVIEWS_DIR = None
        self.modelviews = []
        if (not self.STATICVIEWS_DIR):
            self.populate()
                 
    def populate(self):
        if (not(hasattr(dsettings, 'STATICVIEWS_DIR'))):
            raise ImproperlyConfigured('The static_models app requires a setting STATICVIEWS_DIR to be defined.')
        self.staticviews_dir = dsettings.STATICVIEWS_DIR
            
        if (hasattr(dsettings, 'STATIC_VIEWS')):
            viewsettings = dsettings.STATIC_VIEWS
            self.modelviews = self.build_viewsettings(viewsettings)

    def check_viewsetting(self, vs):
        #if (not ('data' in vs or 'urls' in vs)):
        #    raise ImproperlyConfigured(f'The static_models app requires settings STATIC_MODELVIEWS entires to have a "data" or "urls" key (the key can be empty). entry:{vs}')
        #if (('data' in vs) and not (isinstance(vs['data'], dict))):
        #    raise ImproperlyConfigured(f'The static_models app setting STATIC_MODELVIEWS "data" key is not a dict. entry:{vs}')
        if (not ('urls') and not ('view' in vs)):
            raise ImproperlyConfigured(f'The static_models app requires settings STATIC_MODELVIEWS entriies to have a "view" pt "urls" key. entry:{vs}')
                         
                         
    def build_viewsettings(self, viewsettings):
        b = []
        for mv in viewsettings:
            self.check_viewsetting(mv)

            # test existance of a view?
            # in management or generator?
            if ('view' in mv):
                viewname = mv['view']
                try:
                    view = module_loading.import_string(viewname)
                except ImportError:
                    #print(repr(view))
                    raise ImproperlyConfigured(f'View not visible to a module loader. entry:{viewname}')
                    
            b.append(mv)
        return b
            
    def __repr__(self):
        return "Settings(staticviews_dir:{}, modelviews:{})".format(
            self.staticviews_dir,
            self.modelviews,
        )

# set one up
settings = Settings()



import unittest
from django.test import TestCase
from static_models.utils import ModelGenerator


# ./manage.py test static_models.tests.test_generate
class TestGenerate(TestCase):
    
    def setUp(self):
        #! need to create a model to work on
        
        # self.generator = ModelGenerator(
            # model,
            # view,
            # storage=FileSystemStorage,
            # overwrite=True,
            # path_includes_view=False,
            # slug_in_path=False,
            # extension=None,
        # )
        pass



import os
from django.http.response import FileResponse
from django.core.exceptions import ImproperlyConfigured


class AssuredFileResponse(FileResponse):
    """
    A streaming HTTP response class optimized for files.
    Differs from FileResponse because it assumes the file is what the 
    respose has been told it is, as content_type. No guesses. So you
    would have to set, for example 'application/gzip' for zipped files.
    """
    def __init__(self, *args, content_type=None, **kwargs):
        if content_type is None:
            raise ImproperlyConfigured(
                "AssuredFileResponse requires an attribute 'content_type'"
                " passed in initial parameters")
        super().__init__(*args, content_type=content_type, **kwargs)

    def set_headers(self, filelike):
        """
        Set some common response headers (Content-Length, Content-Type, and
        Content-Disposition) based on the `filelike` response content.
        """
        # encoding_map = {
            # 'bzip2': 'application/x-bzip',
            # 'gzip': 'application/gzip',
            # 'xz': 'application/x-xz',
        # }
        filename = getattr(filelike, 'name', None)
        filename = filename if (isinstance(filename, str) and filename) else self.filename
        if os.path.isabs(filename):
            self['Content-Length'] = os.path.getsize(filelike.name)
        elif hasattr(filelike, 'getbuffer'):
            self['Content-Length'] = filelike.getbuffer().nbytes

        # if self.get('Content-Type', '').startswith('text/html'):
            # if filename:
                # content_type, encoding = mimetypes.guess_type(filename)
                # # Encoding isn't set to prevent browsers from automatically
                # # uncompressing files.
                # content_type = encoding_map.get(encoding, content_type)
                # self['Content-Type'] = content_type or 'application/octet-stream'
            # else:
                # self['Content-Type'] = 'application/octet-stream'
                        
        filename = self.filename or os.path.basename(filename)
        if filename:
            disposition = 'attachment' if self.as_attachment else 'inline'
            try:
                filename.encode('ascii')
                file_expr = 'filename="{}"'.format(filename)
            except UnicodeEncodeError:
                file_expr = "filename*=utf-8''{}".format(quote(filename))
            self['Content-Disposition'] = '{}; {}'.format(disposition, file_expr)
        elif self.as_attachment:
            self['Content-Disposition'] = 'attachment'
            




from django.core.management.base import BaseCommand, CommandError
from static_models.view_generator import ViewStaticManager
#from static_models.settings import settings
from django.conf import settings




class Command(BaseCommand):
    help = 'Create/update static files'
        
    def add_arguments(self, parser):
        parser.add_argument(
            '--sync',
            action='store_true',
            help="Before generation, delete existing files.",
        )
        parser.add_argument(
            '-o',
            '--overwrite',
            action='store_true',
            help="Replace currently existing files (default is to ignore)",
        )
        parser.add_argument(
            '-e',
            '--html_extension',
            action='store_true',
            help="Add '.html' extension to generated files.",
        )

    def normalize_viewsetting(self, vs):
        if (not('view' in vs)):
            vs['view'] = None
        if (not('query' in vs)):
            vs['query'] = None
        if (not('urls' in vs)):
            vs['urls'] = []
        if (not('filename' in vs)):
            vs['filename'] = None
        if (not('filename_from_attribute' in vs)):
            vs['filename_from_attribute'] = 'pk'
        if (not('filepath' in vs)):
            vs['filepath'] = None
        return vs

                    
    def handle(self, *args, **options):         
        extension = ''
        if (options['html_extension']):
            extension = 'html'

        # first check this
        # This is an error not attepted to default
        try:
            settings.STATICVIEWS_DIR
        except AttributeError:
            raise CommandError('The static_models app requires a setting STATICVIEWS_DIR to be defined.')

        valid_entries = []
        for e in settings.STATIC_VIEWS:
            if (not (e.get('urls')) and not (e.get('view'))):
                self.stdout.write(
                    f'WARNING: The static_models app requires settings STATIC_MODELVIEWS entriies to have a "view" pt "urls" key. entry:{e}'
                )
            else:
                valid_entries.append(e)                
        normalised_entries = [self.normalize_viewsetting(e) for e in valid_entries]
            
        # ok, generate
        count = 0

        if (options['sync'] and (options['verbosity'] > 0)):
            self.stdout.write(
                "Info: Directories will be cleared before generation"
            )        
            
        # This is defended to do as much as possible
        for vs in normalised_entries:
            g = None
            try:
                g = ViewStaticManager(
                    vs['view'],
                    vs['query'], 
                    vs['urls'],
                    vs['filename'],
                    vs['filename_from_attribute'],
                    vs['filepath'],
                    overwrite=options['overwrite'],
                    extension=extension,
                )
            except Exception as ex:
                self.stdout.write(
                    f"Fail report: {ex}"
                )  
                self.stdout.write(
                    f"WARNING: Data from a setting failed. setting{vs}"
                )    
            if g:
                delete_count = 0
                if (options['sync']):
                    delete_count = g.delete()
                    
                count = g.create()
                if (options['verbosity'] > 0):
                    print("{} static file(s) created at '{}'".format(count, g.location))



from django.apps import apps
from django.core.exceptions import ImproperlyConfigured


#! unused?
def get_model(model_path):
    Model = None
    if (model_path):
        model_path_elements = model_path.split('.', 1)
        if (len(model_path_elements) != 2):
            raise ImproperlyConfigured("Unable to parse model path (must be form 'app.model'): '{}'".format(model_path))        
        an = model_path_elements[0]
        mn = model_path_elements[1]
        try:
            Model = apps.get_model(an, mn)
        except Exception as e:
            raise ImproperlyConfigured("Unable to locate model from path: '{}'".format(model_path))
    return Model



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['static_models']

package_data = \
{'': ['*'],
 'static_models': ['management/*', 'management/commands/*', 'tests/*']}

setup(name='django-static-models',
      version='1.0.0',
      description='Generate static pages from views',
      author='robert crowther',
      author_email='rw.crowther@gmail.com',
      url='https://github.com/rcrowther/django-static_models',
      packages=packages,
      package_data=package_data,
     )



import posixpath
from pathlib import Path
from django.core.exceptions import ImproperlyConfigured
from django.views.generic.base import View
from django.http import (
    Http404, HttpResponseNotModified,
)
from django.utils._os import safe_join
from django.utils.translation import gettext as _, gettext_lazy
from static_models.response import AssuredFileResponse
from django.views.static import was_modified_since
from django.utils.http import http_date



#! Probably needs a alternate id field slector
class StaticResponseMixin:
    #NB this is a muddle of DetailView, other base view code, and 
    # impatience with Django twaddle. 
    path_root=None

    def __init__(self, **kwargs):   
        kwargs['content_type'] ='text/html' 
        super().__init__(**kwargs)

        #NB TemplateResponseMixin uses late call to check path_root not None.
        # Flexible, but weedy. Define by init, or error.
        if self.path_root is None:
            raise ImproperlyConfigured(
                "StaticView requires an attribute 'path_root'"
                " which can be defined, or passed in initial parameters")

    def get_id_path(self):
        id_path=None
        pk = self.kwargs.get('pk')
        slug = self.kwargs.get('slug')
        path = self.kwargs.get('path')
        
        # start with pk
        if (pk is not None):
            id_path = str(pk)

        # try slug.
        if (slug is not None and (pk is None)):
            id_path = str(slug)

        # try path
        if ((pk is None) and (slug is None) and path is not None):
            id_path = str(path)
                    
        # If nothing defined, error.
        if (not id_path):
            raise AttributeError(
                "StaticView %s must be called with a "
                "parameter named 'pk', 'slug' or 'path' in the URLconf." % self.__class__.__name__
            )            
        return id_path #+ '.html'

    def get_fullpath(self):
        #NB split out so you could, say, implement a 'delete' method on 
        # a view
        return Path(safe_join(self.path_root + self.get_id_path()))
    
    def render_to_response(self, **response_kwargs):
        '''
        Use class data to make a Django HTTP response.
        Would only in most universes be implemented on 'get'
        '''
        #NB some code here comes from django.views.static
        fullpath = self.get_fullpath()
        
        # Respect the If-Modified-Since header.
        statobj = fullpath.stat()
        if not was_modified_since(self.request.META.get('HTTP_IF_MODIFIED_SINCE'),
                                  statobj.st_mtime, statobj.st_size):
            return HttpResponseNotModified()
        try:
            f = fullpath.open('rb')
        except FileNotFoundError:
            raise Http404(_('“%(path)s” does not exist') % {'path': fullpath})

        response = AssuredFileResponse(
            f,
            as_attachment=False,
            content_type='text/html',
            **response_kwargs
        )
        response["Last-Modified"] = http_date(statobj.st_mtime)
        return response
            
            
            
class StaticView(StaticResponseMixin, View):
    '''
    Render a static file.
    Acts on URL parameters 'pk' or 'slug'.
    
    path_root
        Path to a directory of files used as response data. You may want
        to set with settings.MEDIA_ROOT, but this class has no opinion.
    '''       
    def get(self, request, *args, **kwargs):
        return self.render_to_response()



import os
import re
import pathlib
from django.conf import settings
from django.http.request import HttpRequest, QueryDict
from django.core.files.base import ContentFile
from django.db import models
from django.core.exceptions import ImproperlyConfigured
from django.utils import module_loading
from pathlib import Path
from urllib.parse import urlparse


from django.core.handlers.wsgi import WSGIRequest
from io import BytesIO
from urllib.parse import unquote_to_bytes, urljoin, urlparse, urlsplit
from django.utils.encoding import force_bytes
from django.utils.http import urlencode
from django.core.servers.basehttp  import get_internal_wsgi_application


    
                    
class RequestFactory:
    # Build a mocked up request
    # hacked out of django.test.clients
    def __init__(self):
        #self.json_encoder = json_encoder
        #self.defaults = defaults
        #self.cookies = SimpleCookie()
        self.errors = BytesIO()
        
    def _base_environ(self, **request):
        """
        The base environment for a request.
        """
        # This is a minimal valid WSGI environ dictionary, plus:
        # - HTTP_COOKIE: for cookie support,
        # - REMOTE_ADDR: often useful, see #8551.
        # See https://www.python.org/dev/peps/pep-3333/#environ-variables
        return {
            'HTTP_COOKIE': '',
            'PATH_INFO': '/',
            'REMOTE_ADDR': '127.0.0.1',
            'REQUEST_METHOD': 'GET',
            'SCRIPT_NAME': '',
            # Pretend on localhost, like the development it is
            'SERVER_NAME': 'localhost',
            'SERVER_PORT': '80',
            'SERVER_PROTOCOL': 'HTTP/1.1',
            'wsgi.version': (1, 0),
            'wsgi.url_scheme': 'http',
            #'wsgi.input': FakePayload(b''),
            'wsgi.input':            BytesIO(),
            'wsgi.errors': self.errors,
            'wsgi.multiprocess': True,
            'wsgi.multithread': False,
            'wsgi.run_once': False,
            #**self.defaults,
            **request,
        }

    def request(self, **request):
        "Construct a generic request object."
        return WSGIRequest(self._base_environ(**request))

    def generic(self, method, path, data='',
                content_type='application/octet-stream', secure=False,
                **extra):
        """Construct an arbitrary HTTP request."""
        parsed = urlparse(str(path))  # path can be lazy
        data = force_bytes(data, settings.DEFAULT_CHARSET)
        r = {
            'PATH_INFO': self._get_path(parsed),
            'REQUEST_METHOD': method,
            'SERVER_PORT': '443' if secure else '80',
            'wsgi.url_scheme': 'https' if secure else 'http',
        }
        # if data:
            # r.update({
                # 'CONTENT_LENGTH': str(len(data)),
                # 'CONTENT_TYPE': content_type,
                # 'wsgi.input': FakePayload(data),
            # })
        ##r.update(extra)
        # If QUERY_STRING is absent or empty, we want to extract it from the URL.
        if not r.get('QUERY_STRING'):
            # WSGI requires latin-1 encoded strings. See get_path_info().
            query_string = parsed[4].encode().decode('iso-8859-1')
            r['QUERY_STRING'] = query_string
        return self.request(**r)

    def _get_path(self, parsed):
        path = parsed.path
        # If there are parameters, add them
        if parsed.params:
            path += ";" + parsed.params
        path = unquote_to_bytes(path)
        # Replace the behavior where non-ASCII values in the WSGI environ are
        # arbitrarily decoded with ISO-8859-1.
        # Refs comment in `get_bytes_from_wsgi()`.
        return path.decode('iso-8859-1')

    def get(self, path, data=None, secure=False, **extra):
        """Construct a GET request."""
        data = {} if data is None else data
        return self.generic('GET', path, secure=secure, **{
            'QUERY_STRING': urlencode(data, doseq=True),
            **extra,
        })
       
def get_view(viewname):
    '''
    Return a view class.
    viewname
        a path and viewname, as string
    '''
    try:
        view = module_loading.import_string(viewname)
    except ImportError:
        raise ImproperlyConfigured(f'View not visible to a module load. name:"{viewname}"')
    return view


                    
class ViewStaticManager():
    '''
    Generate static files from a Model.
    
    If the data is a model it should not be an instance. The code will 
    assume the view is generic and set the model.
    If the data is a list of dicts the code will pass each dict to the 
    view context.  
    In both cases, the id of the object, and so the filename, will be 
    taken from id_fieldname.
    view
        Must be at least a DetailView. Either a string path or the View 
        class (but not an instance).
    query
        For a model. Currently, only 'all' is accepted. 
    urls
        List of IRLs to evoke the View.
    filename
        A name for one-off view evokation.
    filename_from_attribute
        For use on models. The value from this fieldname is used to 
        provide the filename. Most likely targets a slugfield. Can also
        build from zero-attribute callables. Default is 'pk'
    filepath
        Path from the configured rootdir to the output of this class.
        If empty, the pathroot is empty, unless the data is a model,
        in which case it defaults to a modelname and opetional 
        viewname.
    overwrite
        overwrite existing generated files, even if same size. Default 
        is False.
    extension
        Optional extension name. Dot is added automatically.
    '''
    #! reorder
    def __init__(self, 
        view = None,
        query = None,
        urls = [],
        filename = None,
        filename_from_attribute ='pk',
        filepath = None,
        overwrite = False,
        extension = '',
    ):
        if isinstance(view, str): 
            view = get_view(view)
        self.View = view

        self.basepath = Path(settings.STATICVIEWS_DIR)

        # decide a path from the configured dir to this view output
        self.filepath = self.get_filepath(
            self.basepath,
            filepath, 
            self.View
        )
        
        # make the destination dirs
        self.filepath.mkdir(exist_ok=True)
        
        # set extension
        if (extension):
            extension = '.' + extension        
        self.extension = extension
                      
        self.query = query
        self.urls = urls
        if (urls):
            # If there are URLs to process, start an internal server.
            # This server is like the one run by ''runserver'
            # Radically this is a BaseHandler, set with the local 
            # environment. AFAIK, it needs no closure. Also, errors
            # it throws are natural and good as they read.
            self.server = get_internal_wsgi_application()
        
        self.overwrite = overwrite
        self.filename = filename
        self.filename_from_attribute = filename_from_attribute
        self.id_fieldname = filename_from_attribute


                    
    def get_filepath(self, basedirpath, filepath, view):
        '''
        return 
        The path if given. If not, and a view model exists, use the 
        model name. If the view is not a model, return a default of no 
        path 
        ''' 
        if ((filepath is None) and hasattr(view, 'model')):
            modelname = view.model._meta.model_name
            filepath = self.camelcase_to_underscore(modelname)
        if (filepath is None):
            filepath = ''
        return basedirpath / filepath

    def mk_request(self, url=''):
        request = HttpRequest()
        
        #! need to minimally populate this       
        #- how fragment/query into GET?
        urlparts = urlparse(url)
        querystring = urlparts.query #+ urlparts.fragment
        request.GET = QueryDict(querystring)
        request.path = urlparts.path
        request.method = 'GET'
        return request         
        
    @property
    def location(self):
        '''
        Return a human readable location for the files.
        For display, not good for code.
        return
            The path relative to django base, plus whatever further path 
            decided.
        '''
        return self.filepath.relative_to(self.basepath.parent)
                
    def camelcase_to_underscore(self, s):
        # A utility for generating a filename from view classes
        # Lifted from someplace in Django?
        # https://djangosnippets.org/snippets/585/
        return re.sub('(((?<=[a-z])[A-Z])|([A-Z](?![A-Z]|$)))', '_\\1', s).lower().strip('_')

####
    def render_no_input(self, View, filename):
        if (not(filename)):
            raise ImproperlyConfigured('If no "query" or "url" attribute, then a filename is required')

        # could only be one file?
        full_filepath = self.filepath / (filename + self.extension)
        
        # make a view
        view = View()

        # can be a boring generic request for now.
        view.request = self.mk_request()

        # get a context
        # nearly all views have a context, but the base doesn't. So...
        ctx = {}
        if (hasattr(view, 'get_context_data')):
            ctx = view.get_context_data()

        # get and render response
        count = 0
        response = view.render_to_response(ctx)        
        r = self.render(response, full_filepath)
        if(r):
            count = 1
        return count

####
    def filename_from_url(self, url):
        # Attempt to make a filename from a URL
        # Try for a query, a fragment, then last part of path 
        urlparts = urlparse(url)
        fid = ""
        if (urlparts.query or urlparts.fragment):
            query = ''
            if (urlparts.query):
                query = '?' + urlparts.query
            fragment = ''
            if (urlparts.fragment):
                fragment = '#' + urlparts.fragment
            fid = query + fragment
        if not(fid):
            #! Not good enough. Different no last slash/slash
            path = urlparts.path
            if (path[-1] == "/"):
                path = path[0 : -1]
            path = path.rsplit('/', 1)
            if (path):
                fid = path[-1]
        if not(fid):
            raise ImproperlyConfigured(f'Cannot construct an identifier from given URL. name:"{url}"')
        return fid

    # At some point, angling for this to be the method for view-writing 
    # also
    def writeContent(self, content, full_filepath):
        isfile = full_filepath.is_file()         
        size = 0
        if(isfile):
            size = os.path.getsize(full_filepath)
        generated_size = len(content)
        if ((generated_size != size) or self.overwrite):
            with open(full_filepath, 'wb') as fd:  
                fd.write(content) 
            return full_filepath
        return None
        
    def create_from_url(self, url_filename):
        url = url_filename[0]
        fid = url_filename[1]
        if (fid is None):
            fid = self.filename_from_url(url)
            
        # make the full filepath
        full_filepath = self.filepath / (fid + self.extension)

        req = self.mk_request(url)
        
        # Generate a mock request
        rf = RequestFactory()
        wsgiRequest = rf.get(url)

        # For a response, run through the server
        httpResponse = self.server.get_response(wsgiRequest)
        
        # get content and write to file
        content = httpResponse.content
        return self.writeContent(content, full_filepath)

    def render_urls(self, urls):
        count = 0
        for url_filename in urls.items():
            r = self.create_from_url(url_filename)
            if (r):
                count += 1 
        return count

    def render_query_set(self, View, query='all'):
        view = View()
        if (query == 'all'):
            qs = view.model.objects.all()
        else:
            qs = view.model.objects.filter(query)
            
        count = 0
        for obj in qs: 
            r = self.create_from_obj(view, obj)
            if (r):
                count += 1 
        return count
                
    def create_from_obj(self, view, obj):
        fid = getattr(obj, self.filename_from_attribute) 
        if callable(fid):
            fid = fid()
        full_filepath = self.filepath / (fid + self.extension)
        view.object = obj

        # can be a boring generic request for now.
        view.request = self.mk_request()
        ctx = view.get_context_data()
        response = view.render_to_response(ctx)
        return self.render(response, full_filepath)
        
    def render(self, response, full_filepath):
        '''
        Resder the response o a file
        only if sizes match, or override is set.
        '''
        isfile = full_filepath.is_file() 
        
        size = 0
        if(isfile):
            size = os.path.getsize(full_filepath)
        
        # Response only caches data, resolve
        rr = response.render()
        
        # Gets the data itself
        #! still bytes
        #! where stripped?
        cf = ContentFile(rr.content, name=None)
        generated_size = cf.size
        if ((generated_size != size) or self.overwrite):
            with open(full_filepath, 'wb') as fd:  
                fd.writelines(cf.open())  
            return full_filepath
        return None

    def create(self):
        if (self.urls):
            return self.render_urls(self.urls)
        if (self.query):
            return self.render_query_set(self.View, self.query)
        return self.render_no_input(self.View, self.filename)

    def delete(self):
        '''
        Delete existing generated files.
        Crude, simply clears the target directory. Will not clear the
        directory itself, nor subdirectories.
        '''
        count = 0
        for path in self.filepath.iterdir():
            # The app can generate a nested directory structure. IT is
            # possible a general delete will find a nested directory.
            # We do not want to delete that, as it may not be requested
            # by the user. So gently as psssible, exceptions are quietly 
            # ignored
            r = None
            try:
                r = path.unlink(missing_ok=True)
            except:
                pass
            if (r):
                count += 1
        return count
        
        

        
def file_static_merge(obj, view):
    '''
    Preconfigured Filesystem save.
    PK for id and overwriting.
    '''
    g = ViewStaticManager(obj._meta.model, view, overwrite=True)
    g.obj_create(obj)

def file_static_delete(obj, view):
    '''
    Preconfigured Filesystem delete.
    PK for id.
    '''
    g = ViewStaticManager(obj._meta.model, view)
    g.delete(obj)
